package BLAST::M9Hit;
BEGIN{ push(@INC, $ENV{LOCAL_PERL_MODULES}); }
use Carp;
use warnings;
use strict;
use Sequence::Locus;
use BLAST::M9Statistics;


###################################################
sub new {
    my $pkg = shift;
    bless {
	queryLocus    => Sequence::Locus->new(), 
	subjectLocus  => Sequence::Locus->new(), 
	stats         => BLAST::M9Statistics->new()
    }, $pkg;
}


###################################################
sub setAll {
    my $self = shift;
    my $m9Line = shift or croak "M9 line is not defined.";
    my @cols = split("\t", $m9Line);
    my ($queryID, $subjectID, 
	$pIdentity, $alnLength, 
	$numMismatches, $numGapOpens, 
	$queryStart, $queryEnd, 
	$subjectStart, $subjectEnd, 
	$eValue, $bitScore ) = @cols;
    $bitScore =~ s/\s+//g;
    $self->setQueryLocus( getLocus($queryID, $queryStart, $queryEnd) );
    $self->setSubjectLocus( getLocus($subjectID, $subjectStart, $subjectEnd) );
    $self->setStatistics( $pIdentity, $alnLength, $numMismatches, 
			  $numGapOpens, $eValue, $bitScore );
}


###################################################
sub setQueryLocus {
    my $self = shift;
    my $locus = shift;
    $self->{queryLocus} = $locus;
}


###################################################
sub setSubjectLocus {
    my $self = shift;
    my $locus = shift;
    $self->{subjectLocus} = $locus;
}


###################################################
sub getLocus {
    my $seqID = shift;
    my $start = shift;
    my $end = shift;
    my $locus = Sequence::Locus->new();
    my $strand = '+';
    if( $start > $end ){
	($start, $end) = ($end, $start);
	$strand = '-';
    }
    $locus->setSeqID($seqID);
    $locus->setStart($start);
    $locus->setEnd($end);
    $locus->setStrand($strand);
    return $locus;
}


###################################################
sub setStatistics {
    my $self = shift;
    my $pIdentity = shift;
    my $alnLength = shift;
    my $numMismatches = shift;
    my $numGapOpens = shift;
    my $eValue = shift;
    my $bitScore = shift;
    $self->getStatistics()->setPercentIdentity( $pIdentity );
    $self->getStatistics()->setAlignmentLength( $alnLength );
    $self->getStatistics()->setNumMismatches( $numMismatches );
    $self->getStatistics()->setNumGapOpens( $numGapOpens );
    $self->getStatistics()->setEValue( $eValue );
    $self->getStatistics()->setBitScore( $bitScore );
}


###################################################
sub getQueryLocus {
    my $self = shift;
    return $self->{queryLocus};
}

###################################################
sub getQueryLength {
    my $self = shift;
    return $self->getQueryLocus()->getLength();
}

###################################################
sub getQueryID {
    my $self = shift;
    return $self->getQueryLocus()->getSeqID();
}


###################################################
sub getQueryStart {
    my $self = shift;
    return $self->getQueryLocus()->getStart();
}


###################################################
sub getQueryEnd {
    my $self = shift;
    return $self->getQueryLocus()->getEnd();
}


###################################################
sub getQueryStrand {
    my $self = shift;
    return $self->getQueryLocus()->getStrand();
}

###################################################
sub getQueryMatchLength {
    my $self = shift;
    return $self->getQueryEnd() - $self->getQueryStart() + 1;
}

###################################################
sub getSubjectLength {
    my $self = shift;
    return $self->getSubjectLocus()->getLength();
}

###################################################
sub getSubjectLocus {
    my $self = shift;
    return $self->{subjectLocus};
}


###################################################
sub getSubjectID {
    my $self = shift;
    return $self->getSubjectLocus()->getSeqID();
}


###################################################
sub getSubjectStart {
    my $self = shift;
    return $self->getSubjectLocus()->getStart();
}


###################################################
sub getSubjectEnd {
    my $self = shift;
    return $self->getSubjectLocus()->getEnd();
}

###################################################
sub getSubjectMatchLength {
    my $self = shift;
    return $self->getSubjectEnd() - $self->getSubjectStart() + 1;
}

###################################################
sub getSubjectStrand {
    my $self = shift;
    return $self->getSubjectLocus()->getStrand();
}


###################################################
sub getStatistics {
    my $self = shift;
    return $self->{stats};
}


###################################################
sub getPercentIdentity {
    my $self = shift;
    return $self->getStatistics()->getPercentIdentity();
}


###################################################
sub getAlignmentLength {
    my $self = shift;
    return $self->getStatistics()->getAlignmentLength();
}


###################################################
sub getNumMismatches {
    my $self = shift;
    return $self->getStatistics()->getNumMismatches();
}


###################################################
sub getNumGapOpens {
    my $self = shift;
    return $self->getStatistics()->getNumGapOpens();
}


###################################################
sub getEValue {
    my $self = shift;
    return $self->getStatistics()->getEValue();
}


###################################################
sub getBitScore {
    my $self = shift;
    return $self->getStatistics()->getBitScore();
}

###################################################
sub getLine {
    my $self = shift;
    return $self->getQueryID() . "\t"
	. $self->getSubjectID();
}

###################################################
sub isSame {
    my $self = shift;
    my $hit = shift;
    return 0 unless $self->getQueryID() eq $hit->getQueryID();
    return 0 unless $self->getSubjectID() eq $hit->getSubjectID();
    return 0 unless $self->getQueryStart() == $hit->getQueryStart();
    return 0 unless $self->getQueryEnd() == $hit->getQueryEnd();
    return 0 unless $self->getSubjectStart() == $hit->getSubjectStart();
    return 0 unless $self->getSubjectEnd() == $hit->getSubjectEnd();
    return 1;
}

###################################################
sub isOverlappedOnSubject {
    my $self = shift;
    my $hit = shift;
    return 0 unless $self->getSubjectID() eq $hit->getSubjectID();
    return $self->getSubjectLocus()->isOverlap( $hit->getSubjectLocus() );
}


1;

